from datetime import timedelta

from amino.tc.base import TypeClass, ImplicitInstances
from amino.tc.monoid import Monoid
from amino import Map
from amino.lazy import lazy


class TimedeltaInstances(ImplicitInstances, tpe=timedelta):

    @lazy
    def _instances(self) -> Map[str, TypeClass]:
        return Map({Monoid: TimeDeltaMonoid()})


class TimeDeltaMonoid(Monoid):

    @property
    def empty(self) -> timedelta:
        return timedelta(seconds=0)

    def combine(self, a: timedelta, b: timedelta) -> timedelta:
        return a + b

__all__ = ('TimedeltaInstances',)



import re
import abc
from typing import Callable, Iterable, TypeVar, Generic
from functools import partial

import amino  # NOQA
from amino.tc.apply import Apply
from amino import I
from amino.tc.base import tc_prop

F = TypeVar('F')
A = TypeVar('A')
B = TypeVar('B')


class FlatMap(Generic[F], Apply[F]):
    _flat_map_re = re.compile('^flat_map(\d+)$')
    _product_re = re.compile('^product(\d+)$')

    def ap(self, fa: F, ff: F):
        f = lambda f: self.map(fa, f)
        return self.flat_map(ff, f)

    @abc.abstractmethod
    def flat_map(self, fa: F, f: Callable[[A], F]) -> F:
        ...

    def __floordiv__(self, fa, f):
        return self.flat_map(fa, f)

    def product(self, fa: F, fb: F) -> F:
        f = lambda a: self.map(fb, lambda b: (a, b))
        return self.flat_map(fa, f)

    __and__ = product

    def __getattr__(self, name):
        flat_map = self._flat_map_re.match(name)
        product = self._product_re.match(name)
        if flat_map is not None:
            return partial(self.flat_map_n, int(flat_map.group(1)))
        elif product is not None:
            return partial(self.product_n, int(product.group(1)))
        else:
            return super().__getattr__(name)

    def flat_map_n(self, num, fa: F, f: Callable[..., F]) -> F:
        def wrapper(args):
            if len(args) != num:
                msg = 'passed {} args to {}.flat_map{}'
                name = self.__class__.__name__
                raise TypeError(msg.format(len(args), name, num))
            return f(*args)
        return self.flat_map(fa, wrapper)

    def product_n(self, num: int, fa: F, *fs: Iterable[F]):
        from amino.list import List
        if len(fs) != num:
            msg = 'passed {} args to {}.product{}'
            name = self.__class__.__name__
            raise TypeError(msg.format(len(fs), name, num))
        def add(a, b):
            return self.flat_map(a, lambda a: self.map(b, lambda b: a + (b,)))
        init = self.map(fa, lambda a: (a,))
        return List.wrap(fs).fold_left(init)(add)

    def flat_pair(self, fa: F, f: Callable[[A], 'amino.maybe.Maybe[B]']) -> F:
        cb = lambda a: f(a).map(lambda b: (a, b))
        return self.flat_map(fa, cb)

    def flat_replace(self, fa: F, fb: F) -> F:
        cb = lambda a: fb
        return self.flat_map(fa, cb)

    @tc_prop
    def join(self, fa: F) -> F:
        return self.flat_map(fa, I)

__all__ = ('FlatMap',)



import abc
from typing import Generic, TypeVar, Callable, Tuple, cast, Type

from amino.tc.base import Implicits
from amino.tc.monad import Monad
from amino.tc.zip import Zip
from amino.instances.list import ListTraverse
from amino import List, Maybe, Either, Eval
from amino.id import Id

S = TypeVar('S')
A = TypeVar('A')
B = TypeVar('B')
G = TypeVar('G')


class StateT(Generic[G, S, A]):
    pass


def state_t(tpe: Type[G]) -> type:
    class F(Generic[A], Implicits):
        pass
    cast(abc.ABCMeta, F).register(tpe)
    monad: Monad = cast(Monad, Monad.fatal(tpe))
    class State(Generic[S, A], StateT[tpe, S, A], Implicits, implicits=True, auto=True):

        @staticmethod
        def apply(f: Callable[[S], F[Tuple[S, A]]]) -> 'State[S, A]':
            return State(monad.pure(f))

        @staticmethod
        def applyF(run_f: F[Callable[[S], F[Tuple[S, A]]]]) -> 'State[S, A]':
            return State(run_f)

        @staticmethod
        def inspect(f: Callable[[S], A]) -> 'State[S, A]':
            def g(s: S) -> F[Tuple[S, A]]:
                return monad.pure((s, f(s)))
            return State.apply(g)

        @staticmethod
        def pure(a: A) -> 'State[S, A]':
            return State.apply(lambda s: monad.pure((s, a)))

        @staticmethod
        def modify(f: Callable[[S], S]) -> 'State[S, A]':
            return State.apply(lambda s: monad.pure((f(s), None)))

        @staticmethod
        def modify_f(f: Callable[[S], F[S]]) -> 'State[S, A]':
            return State.apply(lambda s: (f(s).map(lambda a: (a, None))))

        @staticmethod
        def set(s: S) -> 'State[S, A]':
            return State.modify(lambda s0: s)

        @property
        def tpe(self) -> type:
            return tpe

        def __init__(self, run_f: F[Callable[[S], F[Tuple[S, A]]]]) -> None:
            self.run_f = run_f

        def run(self, s: S) -> F[Tuple[S, A]]:
            return self.run_f.flat_map(lambda f: f(s))

        def run_s(self, s: S) -> F[S]:
            return self.run(s).map(lambda a: a[0])

        def run_a(self, s: S) -> F[S]:
            return self.run(s).map(lambda a: a[1])

        def __str__(self) -> str:
            return f'State({self.run_f})'

        def flat_map_f(self, f: Callable[[A], F[B]]) -> 'State[S, B]':
            def h(s: S, a: A) -> F[Tuple[S, B]]:
                return f(a).map(lambda b: (s, b))
            def g(fsa: F[Tuple[S, A]]) -> F[Tuple[S, B]]:
                return fsa.flat_map2(h)
            run_f1 = self.run_f.map(lambda sfsa: lambda a: g(sfsa(a)))
            return State.applyF(run_f1)

        def transform(self, f: Callable[[Tuple[S, A]], Tuple[S, B]]) -> 'State[S, B]':
            def g(fsa: F[Tuple[S, A]]) -> F[Tuple[S, B]]:
                return fsa.map2(f)
            run_f1 = self.run_f.mmap(lambda sfsa: lambda a: g(sfsa(a)))
            return State.applyF(run_f1)
    class StateMonad(Monad, tpe=State):

        def pure(self, a: A) -> State[S, A]:
            return State.pure(a)

        def flat_map(self, fa: State[S, A], f: Callable[[A], State[S, B]]) -> State[S, B]:
            def h(s: S, a: A) -> F[Tuple[S, B]]:
                return f(a).run(s)
            def g(fsa: F[Tuple[S, A]]) -> F[Tuple[S, B]]:
                return fsa.flat_map2(h)
            def i(sfsa: Callable[[S], F[Tuple[S, A]]]) -> Callable[[S], F[Tuple[S, B]]]:
                return lambda a: g(sfsa(a))
            run_f1 = fa.run_f.map(i)
            return State.applyF(run_f1)
    class StateZip(Zip, tpe=State):

        def zip(self, fa: State[S, A], fb: State[S, B], *fs: State) -> State[S, List[A]]:
            return ListTraverse().sequence(List(fa, fb, *fs), State)
    return type(f'State_{tpe.__name__}', (Generic[S, A], State,), dict())


MaybeState = state_t(Maybe)
EitherState = state_t(Either)
EvalState = state_t(Eval)
IdState = state_t(Id)

__all__ = ('StateT', 'state_t', 'MaybeState', 'EitherState', 'EvalState', 'IdState')



import abc
from typing import TypeVar, Generic, Callable, Union, cast

from amino.tc.base import TypeClass, tc_prop
from amino import maybe  # NOQA
from amino.boolean import Boolean

F = TypeVar('F')
A = TypeVar('A')
B = TypeVar('B')


class Optional(Generic[F], TypeClass):

    @abc.abstractmethod
    def to_maybe(self, fa: F) -> 'maybe.Maybe[B]':
        ...

    def get_or_else(self, fa: F, a: Union[A, Callable[[], A]]):
        return self.to_maybe(fa).get_or_else(a)

    @abc.abstractmethod  # type: ignore
    def to_either(self, fb: F, left: Union[A, Callable[[], A]]
                  ) -> 'Either[A, B]':
        ...

    __or__ = get_or_else

    @abc.abstractmethod
    def present(self, fa: F) -> Boolean:
        ...

    def or_else(self, fa: F, a: Union[F, Callable[[], F]]):
        return fa if self.present(fa) else maybe.call_by_name(a)

    o = or_else

    def task(self, fa: F, err: str=''):
        from amino.task import Task
        return Task.from_either(self.to_either(fa, err))

    @tc_prop
    def true(self, fa: F) -> Boolean:
        return self.to_maybe(fa).exists(bool)

    @tc_prop
    def _unsafe_value(self, fa: F) -> A:
        return cast(A, self.to_maybe(fa)._get)

__all__ = ('Optional',)



from functools import wraps, partial
from inspect import getfullargspec
from typing import Callable, Union, Any, TypeVar, Tuple

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


def curried(func):
    @wraps(func)
    def _curried(*args, **kwargs):
        f = func
        count = 0
        while isinstance(f, partial):
            if f.args:
                count += len(f.args)
            f = f.func
        spec = getfullargspec(f)
        if count == len(spec.args) - len(args):
            return func(*args, **kwargs)
        else:
            return curried(partial(func, *args, **kwargs))
    return _curried


class Identity:

    def __init__(self) -> None:
        self.__name__ = 'identity'

    def __call__(self, a):
        return a

    def __str__(self):
        return '(a => a)'

I = Identity()


class Val:

    def __init__(self, value) -> None:
        self.value = value
        self.__name__ = self.__class__.__name__

    def __call__(self):
        return self.value

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.value)


class ReplaceVal(Val):

    def __call__(self, *a, **kw):
        return super().__call__()


def flip(a, b):
    return b, a

CallByName = Union[Any, Callable[[], Any]]


def call_by_name(b: CallByName):
    return b() if callable(b) else b


def is_not_none(a):
    return a is not None


def tupled2(f: Callable[[A, B], C]) -> Callable[[Tuple[A, B]], C]:
    def wrap(a: Tuple[A, B]) -> C:
        return f(a[0], a[1])
    return wrap

__all__ = ('curried', 'I', 'flip', 'call_by_name', 'Val', 'ReplaceVal', 'is_not_none', 'tupled2')



from typing import Callable, Generic, TypeVar

from amino import List, I, _
from amino.lazy import lazy
from amino.tc.base import Implicits
from amino.tc.flat_map import FlatMap

A = TypeVar('A')
B = TypeVar('B')


class F(Generic[A]):
    pass


class Eff(Generic[A], Implicits, implicits=True, auto=True):
    ''' An effect stack transformer.
    Wraps arbitrarily nested effects, like Task[List[Maybe[A]]].
    '''

    def __init__(self, value: F[A], effects: List[type]=List(), depth: int=1) -> None:
        self.value = value
        self.effects = effects
        self.depth = depth

    def __repr__(self):
        return 'Eff[{}]({!r})'.format(self.effects.mk_string(','), self.value)

    def copy(self, value):
        return Eff(value, self.effects, self.depth)

    @lazy
    def all_effects(self):
        return self.effects.cons(type(self.value))

    def _map(self, f: Callable):
        g = List.wrap(range(self.depth)).fold_left(f)(lambda z, i: lambda a: a.map(z))
        return g(self.value)

    def map(self, f: Callable):
        return self.copy(self._map(lambda a: a.map(f)))

    __truediv__ = map

    def _flat_map(self, f: Callable):
        ''' **f** must return the same stack type as **self.value** has.
        Iterates over the effects, sequences the inner instance
        successively to the top and joins with the outer instance.
        Example:
        List(Right(Just(1))) => List(Right(Just(List(Right(Just(5))))))
        => List(List(Right(Just(Right(Just(5))))))
        => List(Right(Just(Right(Just(5)))))
        => List(Right(Right(Just(Just(5)))))
        => List(Right(Just(Just(5))))
        => List(Right(Just(5)))
        Note: Task works only as outermost effect, as it cannot sequence
        '''
        index = List.range(self.depth + 1)
        g = index.fold_left(f)(lambda z, i: lambda a: a.map(z))
        nested = g(self.value)
        def sequence_level(z, depth, tpe):
            nesting = lambda z, i: lambda a: a.map(z).sequence(tpe)
            lifter = List.range(depth).fold_left(I)(nesting)
            return z // lifter
        def sequence_type(z, data):
            return lambda a: sequence_level(a, *data).map(z)
        h = self.all_effects.reversed.with_index.fold_left(I)(sequence_type)
        return h(nested)

    def flat_map(self, f: Callable):
        ''' Calls **f** with the inner values, if present, and sequences
        the result, which is assumed to be a complete stack, one by one
        through the old stack.
        '''
        return self.copy(self._flat_map(f))

    __floordiv__ = flat_map

    def flat_map_inner(self, f: Callable):
        return self.copy(self.value.map(lambda a: a.flat_map(f)))


class EffFlatMap(FlatMap, tpe=Eff):

    def map(self, fa: Eff[A], f: Callable[[A], B]) -> Eff[B]:
        return Eff.map(fa, f)

    def flat_map(self, fa: Eff[A], f: Callable[[A], Eff[B]]) -> Eff[B]:
        return Eff.flat_map(fa, f)

__all__ = ('Eff',)



from typing import TypeVar, Callable, Any, Generator, cast, Optional
import functools

from amino.tc.base import Implicits

F = TypeVar('F', bound=Implicits)
A = TypeVar('A')


def do(f: Callable[..., Generator[F, A, F]]) -> Callable:
    @functools.wraps(f)
    def do_loop(*a: Any, **kw: Any) -> F:
        itr = f(*a, **kw)
        c: Optional[F] = None
        def send(val: A) -> F:
            nonlocal c
            try:
                c = itr.send(val)
                return c.flat_map(send)
            except StopIteration:
                return cast(F, c)
        return send(cast(A, None))
    return do_loop

__all__ = ('do',)



import abc
from typing import TypeVar, Generic

from amino.tc.base import TypeClass

A = TypeVar('A')


class Monoid(Generic[A], TypeClass):

    @abc.abstractproperty
    def empty(self):
        ...

    @abc.abstractmethod
    def combine(self, a, b):
        ...

__all__ = ('Monoid',)



import re
import sys
import logging
import operator
from pathlib import Path
from logging import LogRecord
from typing import Callable, Any, Union, cast

from amino.lazy import lazy

import amino
import amino.maybe
from amino.options import EnvOption

VERBOSE = 15
TEST = 12
DEBUG1 = DDEBUG = 7
DEBUG2 = 4
logging.addLevelName(VERBOSE, 'VERBOSE')
logging.addLevelName(TEST, 'TEST')
logging.addLevelName(DEBUG1, 'DEBUG1')
logging.addLevelName(DEBUG2, 'DEBUG2')


def seq_to_str(data: Union[str, 'amino.List', Any]) -> str:
    return cast(amino.List, data).join_lines if isinstance(data, amino.List) else str(data)


class LazyRecord(logging.LogRecord):

    def __init__(self, name, level, pathname, lineno, msg, args, exc_info, func=None, sinfo=None,  # type: ignore
                 **kwargs) -> None:
        super().__init__(name, level, pathname, lineno, msg, args, exc_info, func=func, sinfo=sinfo)
        self._data = msg
        self._args = args

    @lazy
    def _cons_message(self) -> str:
        data = self._data(*self._args) if callable(self._data) else self._data
        return seq_to_str(data)

    def getMessage(self) -> str:
        return self._cons_message if self.levelname in ('DEBUG1', 'DEBUG2') else super().getMessage()


class Logger(logging.Logger):

    def test(self, message, *args, **kws):
        if self.isEnabledFor(TEST):
            self._log(TEST, message, args, **kws)

    def verbose(self, message, *args, **kws):
        if self.isEnabledFor(VERBOSE):
            self._log(VERBOSE, message, args, **kws)

    def debug1(self, f: Callable[..., str], *args: Any) -> None:
        if self.isEnabledFor(DEBUG1):
            self._log(DEBUG1, f, args)  # type: ignore

    def debug2(self, f: Callable[..., str], *args: Any) -> None:
        if self.isEnabledFor(DEBUG2):
            self._log(DEBUG2, f, args)  # type: ignore

    ddebug = debug2

    def caught_exception(self, when: str, exc: Exception, *a: Any, **kw: Any) -> None:
        headline = 'caught exception while {}:'.format(when)
        self.exception(headline, exc_info=(type(exc), exc, exc.__traceback__))

    def makeRecord(self, name: str, level: int, fn: str, lno: int, msg: Any, args: Any, exc_info: Any, func: Any=None,
                   extra: Any=None, sinfo: Any=None) -> LogRecord:
        return LazyRecord(name, level, fn, lno, msg, args, exc_info, func, sinfo)


def install_logger_class() -> None:
    logging.setLoggerClass(Logger)

install_logger_class()

log = logging.getLogger('amino')
amino_root_logger = log
log.setLevel(DEBUG2)
log.propagate = False


def amino_logger(name: str) -> logging.Logger:
    return amino_root_logger.getChild(name)

_stdout_logging_initialized = False

env_log_level = EnvOption('AMINO_LOG_LEVEL')


def init_loglevel(handler: logging.Handler, level: int=None) -> None:
    (
        amino.Maybe.check(level)
        .o(env_log_level.value)
        .o(amino.Boolean(amino.development).flat_m(VERBOSE)) %
        handler.setLevel
    )

amino_stdout_handler = logging.StreamHandler(stream=sys.stdout)


def amino_stdout_logging(level: int=None) -> None:
    global _stdout_logging_initialized
    if not _stdout_logging_initialized:
        amino_root_logger.addHandler(amino_stdout_handler)
        init_loglevel(amino_stdout_handler, level)
        _stdout_logging_initialized = True


default_logfile = Path.home() / '.python' / 'log'
_file_fmt = ('{asctime} [{levelname} @ {name}:{funcName}:{lineno}] {message}')


def amino_file_logging(logger: logging.Logger, level: int=logging.DEBUG, logfile: Path=default_logfile,
                       fmt: str=None) -> logging.Handler:
    logfile.parent.mkdir(exist_ok=True)
    formatter = logging.Formatter(fmt or _file_fmt, style='{')
    handler = logging.FileHandler(str(logfile))
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    init_loglevel(handler, level)
    return handler


def amino_root_file_logging(level: int=logging.DEBUG, **kw: Any) -> logging.Handler:
    return amino_file_logging(amino_root_logger, level, **kw)


class Logging:

    @property
    def log(self) -> Logger:
        return self._log

    @lazy
    def _log(self) -> Logger:
        return amino_logger(self.__class__.__name__)

    def _p(self, a):
        v = self.log.verbose
        v(a)
        return a

    def _dbg(self, fmt, level=DEBUG2):
        def log(a):
            msg = fmt.format(a)
            self.log.log(level, msg)
            return a
        return log


def sub_loggers(loggers, root):
    from amino import Map, _, L
    children = loggers.keyfilter(L(re.match)('{}\.[^.]+$'.format(root), _))
    sub = (children.k / L(sub_loggers)(loggers, _)).fold_left(Map())(operator.pow)
    return Map({loggers[root]: sub})


def logger_tree(root):
    from amino import __, Map
    m = Map(logging.Logger.manager.loggerDict)
    all = m.keyfilter(__.startswith(root))
    return sub_loggers(all, 'amino')


def indent(strings: 'amino.List[str]', level: int, width: int=1) -> 'amino.List[str]':
    ws = ' ' * level * width
    return strings.map(str).map(ws.__add__)


def format_logger_tree(tree: 'amino.Map[Logger, Any]', fmt_logger: Callable[[Logger], str], level: int=0
                       ) -> 'amino.List[str]':
    from amino import _, L
    sub_f = L(format_logger_tree)(_, fmt_logger, level=level + 1)
    formatted = tree.bimap(fmt_logger, sub_f)
    return indent(formatted.flat_map2(lambda a, b: b.cons(a)), level)


def print_log_info(out: Callable[[str], None]) -> None:
    lname = lambda l: logging.getLevelName(l.getEffectiveLevel())
    hlname = lambda h: logging.getLevelName(h.level)
    def handler(h: logging.Handler) -> str:
        return '{}({})'.format(h.__class__.__name__, hlname(h))
    def logger(l: logging.Logger) -> str:
        handlers = ','.join(list(map(handler, l.handlers)))
        return '{}: {} {}'.format(l.name, lname(l), handlers)
    out(format_logger_tree(logger_tree('amino'), logger).join_lines)
    out('-------')
    out(str(env_log_level))
    out(str(amino.options.development))

__all__ = ('amino_root_logger', 'amino_stdout_logging', 'amino_file_logging', 'amino_root_file_logging',
           'print_log_info')



import operator
from typing import Any, Callable

from amino.util.fun import format_funcall


def lop(op, s):
    def oper(self, a):
        return self.__lop__(op, s, a)
    return oper


def rop(op, s):
    def oper(self, a):
        return self.__rop__(op, s, a)
    return oper


class Opers:

    __add__ = lop(operator.add, "+")
    __mul__ = lop(operator.mul, "*")
    __sub__ = lop(operator.sub, "-")
    __mod__ = lop(operator.mod, "%")
    __pow__ = lop(operator.pow, "**")
    __and__ = lop(operator.and_, "&")
    __or__ = lop(operator.or_, "|")
    __xor__ = lop(operator.xor, "^")
    __div__ = lop(operator.truediv, "/")
    __divmod__ = lop(divmod, "/")
    __floordiv__ = lop(operator.floordiv, "/")
    __truediv__ = lop(operator.truediv, "/")
    __lshift__ = lop(operator.lshift, "<<")
    __rshift__ = lop(operator.rshift, ">>")
    __lt__ = lop(operator.lt, "<")
    __le__ = lop(operator.le, "<=")
    __gt__ = lop(operator.gt, ">")
    __ge__ = lop(operator.ge, ">=")
    __eq__ = lop(operator.eq, "==")
    __ne__ = lop(operator.ne, "!=")
    __radd__ = rop(operator.add, "+")
    __rmul__ = rop(operator.mul, "*")
    __rsub__ = rop(operator.sub, "-")
    __rmod__ = rop(operator.mod, "%")
    __rpow__ = rop(operator.pow, "**")
    __rdiv__ = rop(operator.truediv, "/")
    __rdivmod__ = rop(divmod, "/")
    __rtruediv__ = rop(operator.truediv, "/")
    __rfloordiv__ = rop(operator.floordiv, "/")
    __rlshift__ = rop(operator.lshift, "<<")
    __rrshift__ = rop(operator.rshift, ">>")
    __rand__ = rop(operator.and_, "&")
    __ror__ = rop(operator.or_, "|")
    __rxor__ = rop(operator.xor, "^")


class Anon:

    def __eval__(self) -> Callable:
        return eval(repr(self))


class AnonChain:

    def __init__(self, pre, post, pre_count) -> None:
        self.__pre = pre
        self.__post = post
        self.__pre_count = pre_count
        self.__name__ = self.__pre.__name__

    def __call__(self, *a, **kw):
        pre_args, post_args = a[:self.__pre_count], a[self.__pre_count:]
        pre_result = self.__pre(*pre_args)
        return self.__post(pre_result, *post_args)

    def __str__(self):
        return '({} >> {})'.format(self.__pre, self.__post)


class ChainedAnonMethodCall(Anon, Opers):

    def __init__(self, pre: 'AnonMethodCall', post: str, args: tuple, kw: dict) -> None:
        self.__pre = pre
        self.__post = post
        self.__args = args
        self.__kw = kw

    def __repr__(self) -> Callable:
        return f'lambda a: {self.__post}'

    def __call__(self, *a: Any, **kw: Any) -> Any:
        return self.__eval__()(self.__pre(*a, **kw))(*self.__args, **self.__kw)

    def __lop__(self, op, s, a):
        return ChainedAnonMethodCall(self, f'(lambda b: a {s} b)', (a,), {})

    def __rop__(self, op, s, a):
        return ChainedAnonMethodCall(self, f'(lambda b: b {s} a)', (a,), {})

    def __getattr__(self, name: str) -> 'MethodChain':
        return MethodChain(self, f'a.{name}')


class MethodChain:

    def __init__(self, pre: 'AnonMethodCall', post: str) -> None:
        self.__pre = pre
        self.__post = post

    def __call__(self, *a: Any, **kw: Any) -> ChainedAnonMethodCall:
        return ChainedAnonMethodCall(self.__pre, self.__post, a, kw)

    def __getattr__(self, name: str) -> 'MethodChain':
        return MethodChain(self.__pre, f'{self.__post}.{name}')


class AnonMethodCall(Anon, Opers):

    def __init__(self, pre: str, args: tuple, kw: dict) -> None:
        self.__pre = pre
        self.__args = args
        self.__kw = kw

    def __repr__(self) -> str:
        return f'lambda a: {self.__pre}'

    def __call__(self, *a: Any, **kw: Any) -> Any:
        return self.__eval__()(*a, **kw)(*self.__args, **self.__kw)

    def __getattr__(self, name: str) -> MethodChain:
        return MethodChain(self, f'a.{name}')

    def __getitem__(self, key: Any) -> ChainedAnonMethodCall:
        return ChainedAnonMethodCall(self, f'a.__getitem__', (key,), {})

    def __str__(self) -> str:
        return f'lambda a: {format_funcall(self.__pre, self.__args, self.__kw)}'

    def __substitute_object__(self, obj: Any) -> Callable:
        return eval(f'lambda a: {self.__pre}')(obj)(*self.__args, **self.__kw)

    def __lop__(self, op, s, a):
        return ChainedAnonMethodCall(self, f'(lambda b: a {s} b)', (a,), {})

    def __rop__(self, op, s, a):
        return ChainedAnonMethodCall(self, f'(lambda b: b {s} a)', (a,), {})


class MethodRef(Anon):

    def __init__(self, pre: str) -> None:
        self.__pre = pre

    def __repr__(self) -> str:
        return f'lambda a: {self.__pre}'

    def __call__(self, *a: Any, **kw: Any) -> AnonMethodCall:
        return AnonMethodCall(self.__pre, a, kw)

    def __getattr__(self, name: str) -> 'MethodRef':
        return MethodRef(f'{self.__pre}.{name}')

    def __substitute_object__(self, obj: Any) -> Callable:
        return eval(f'lambda a: {self.__pre}')(obj)


class MethodLambda:

    def __getattr__(self, name: str) -> MethodRef:
        return MethodRef(f'a.{name}')

    def __call__(self, *a: Any, **kw: Any) -> AnonMethodCall:
        return MethodRef('a')(*a, **kw)

    def __getitem__(self, key: Any) -> MethodRef:
        return MethodRef('a.__getitem__')(key)

MethodLambdaInst = MethodLambda()


__all__ = ('MethodLambdaInst', 'Opers')



import abc
from typing import Callable, TypeVar, Generic, Union, cast, Any

from amino.logging import Logging
from amino import LazyList, Boolean, __, _, Either, Right, Maybe, Left, L, Map, curried
from amino.boolean import false, true
from amino.tc.base import Implicits
from amino.tc.flat_map import FlatMap
from amino.func import call_by_name
from amino.lazy_list import LazyLists


def indent(strings: LazyList[str]) -> LazyList[str]:
    return strings.map(' ' + _)


Data = TypeVar('Data')
Data1 = TypeVar('Data1')
Sub = TypeVar('Sub')
Sub1 = TypeVar('Sub1')
A = TypeVar('A')
B = TypeVar('B')
Z = TypeVar('Z')
Key = Union[str, int]


class Node(Generic[Data, Sub], Logging, abc.ABC, Implicits, implicits=True, auto=True):

    @abc.abstractproperty
    def sub(self) -> Sub:
        ...

    @abc.abstractproperty
    def sub_l(self) -> LazyList['Node[Data, Any]']:
        ...

    @abc.abstractmethod
    def _strings(self) -> LazyList[str]:
        ...

    @property
    def strings(self) -> LazyList[str]:
        return self._strings()

    def _show(self) -> str:
        return self._strings().mk_string('\n')

    @property
    def show(self) -> str:
        return self._show()

    @abc.abstractmethod
    def foreach(self, f: Callable[['Node'], None]) -> None:
        ...

    @abc.abstractmethod
    def filter(self, pred: Callable[['Node'], bool]) -> 'Node':
        ...

    def filter_not(self, pred: Callable[['Node'], bool]) -> 'Node':
        return self.filter(lambda a: not pred(a))

    @abc.abstractproperty
    def flatten(self) -> 'LazyList[Any]':
        ...

    @abc.abstractmethod
    def contains(self, target: 'Node') -> Boolean:
        ...

    @abc.abstractmethod
    def lift(self, key: Key) -> 'SubTree':
        ...

    def __getitem__(self, key: Key) -> 'SubTree':
        return self.lift(key)

    @abc.abstractproperty
    def s(self) -> 'SubTree':
        ...

    @abc.abstractproperty
    def empty(self) -> Boolean:
        ...

    @curried
    def fold_left(self, z: Z, f: Callable[[Z, 'Node'], Z]) -> Z:
        z1 = f(z, self)
        return self.sub_l.fold_left(z1)(lambda z2, a: a.fold_left(z2)(f))

    @abc.abstractmethod
    def replace(self, data: LazyList['Node[Data1, Sub1]']) -> 'Node[Data1, Sub1]':
        ...

    @abc.abstractmethod
    def map_nodes(self, f: Callable[['Node[Data, Sub]'], 'Node[Data, Sub]']) -> 'Node[Data, Sub]':
        ...


class Inode(Generic[Data, Sub], Node[Data, Sub]):

    @abc.abstractproperty
    def sub(self) -> LazyList[Any]:
        ...

    def foreach(self, f: Callable[[Node], None]) -> None:
        f(self)
        self.sub_l.foreach(__.foreach(f))

    @property
    def flatten(self) -> LazyList[Any]:
        return self.sub_l.flat_map(_.flatten).cons(self)

    def contains(self, target: Node) -> Boolean:
        return self.sub_l.contains(target)

    @property
    def empty(self) -> Boolean:
        return self.data.empty


class ListNode(Generic[Data], Inode[Data, LazyList[Node[Data, Any]]]):

    def __init__(self, sub: LazyList[Node[Data, Any]]) -> None:
        self.data = sub

    @property
    def sub(self) -> LazyList[Node[Data, Any]]:
        return self.data

    @property
    def sub_l(self) -> LazyList[Node[Data, Any]]:
        return self.sub

    @property
    def _desc(self) -> str:
        return '[]'

    def _strings(self) -> LazyList[str]:
        return indent(self.sub // (lambda a: a._strings())).cons(self._desc)

    @property
    def head(self) -> 'SubTree':
        return self.lift(0)

    @property
    def last(self) -> 'SubTree':
        return self.lift(-1)

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.sub.map(str).mk_string(','))

    def __repr__(self) -> str:
        return str(self)

    def lift(self, key: Key) -> 'SubTree':
        return (
            SubTreeInvalid(key, 'ListNode index must be int')
            if isinstance(key, str) else
            self.sub.lift(key) / L(SubTree.cons)(_, key) | (lambda: SubTreeInvalid(key, 'ListNode index oob'))
        )

    def replace(self, sub: LazyList[Any]) -> Node:
        return ListNode(sub)

    def filter(self, pred: Callable[[Node], bool]) -> Node:
        def filt(n: Node) -> bool:
            return (
                pred(n)
                if isinstance(n, LeafNode) else
                not n.empty
            )
        return self.replace(self.sub.map(__.filter(pred)).filter(filt))

    @property
    def s(self) -> 'SubTree':
        return SubTreeList(self, 'root')

    def map_nodes(self, f: Callable[['Node[Data, Sub]'], 'Node[Data, Sub]']) -> 'Node[Data, Sub]':
        return f(ListNode(self.sub.map(lambda a: a.map_nodes(f))))


class MapNode(Generic[Data], Inode[Data, Map[str, Node[Data, Any]]]):

    def __init__(self, data: Map[str, Node[Data, Any]]) -> None:
        self.data = data

    @property
    def sub(self) -> Map[str, Node[Data, Any]]:
        return self.data

    @property
    def sub_l(self) -> LazyList[Node[Data, Any]]:
        return LazyList(self.data.v)

    @property
    def _desc(self) -> str:
        return '{}'

    def _strings(self) -> LazyList[str]:
        return indent(self.sub_l // (lambda a: a._strings())).cons(self._desc)

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.sub_l)

    def __repr__(self) -> str:
        return str(self)

    # TODO allow int indexes into sub_l
    def lift(self, key: Key) -> 'SubTree':
        def err() -> 'SubTree':
            keys = ', '.join(self.data.keys())
            return SubTreeInvalid(key, f'MapNode({self.rule}) invalid key ({keys})')
        return (
            self.data.lift(key) /
            L(SubTree.cons)(_, key) |
            err
        )

    def replace(self, sub: Map[str, Node]) -> Node:
        return MapNode(sub)

    def filter(self, pred: Callable[[Node], bool]) -> Node:
        def filt(n: Node) -> bool:
            return (
                pred(n)
                if isinstance(n, LeafNode) else
                not n.empty
            )
        return self.replace(self.data.valmap(__.filter(pred)).valfilter(filt))

    @property
    def s(self) -> 'SubTree':
        return SubTreeMap(self, 'root')

    def map_nodes(self, f: Callable[['Node[Data, Sub]'], 'Node[Data, Sub]']) -> 'Node[Data, Sub]':
        return f(MapNode(self.sub.valmap(lambda a: a.map_nodes(f))))


class LeafNode(Generic[Data], Node[Data, None]):

    def __init__(self, data: Data) -> None:
        self.data = data

    def _strings(self) -> LazyList[Data]:
        return LazyLists.cons(self.data)

    @property
    def sub(self) -> None:
        pass

    @property
    def sub_l(self) -> LazyList[Node[Data, Any]]:
        return LazyList([])

    def foreach(self, f: Callable[[Node], None]) -> None:
        f(self)

    def filter(self, pred: Callable[[Node], bool]) -> Node:
        return self

    @property
    def flatten(self) -> LazyList[Any]:
        return LazyLists.cons(self)

    def contains(self, target: Node) -> Boolean:
        return false

    def lift(self, key: Key) -> 'SubTree':
        return SubTreeInvalid(key, 'LeafNode cannot be indexed')

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.data)

    def __repr__(self) -> str:
        return str(self)

    @property
    def empty(self) -> Boolean:
        return false

    @property
    def s(self) -> 'SubTree':
        return SubTreeLeaf(self, 'root')

    def replace(self, sub: Data) -> Node:
        return LeafNode(sub)

    def map_nodes(self, f: Callable[['Node[Data, Sub]'], 'Node[Data, Sub]']) -> 'Node[Data, Sub]':
        return f(self)


class TreeFlatMap(FlatMap, tpe=Node):

    def flat_map(self, fa: Node[A, Any], f: Callable[[A], Node[B, Any]]) -> Node[B, Any]:
        return (
            self.flat_map_inode(fa, f)
            if isinstance(fa, Inode) else
            self.flat_map_leaf(fa, f)
        )

    def flat_map_inode(self, fa: Inode[A, Any], f: Callable[[A], Node[B, Any]]) -> Node[B, Any]:
        def err() -> Inode[A, Any]:
            raise Exception(f'invalid sub for `TreeFlatMap.flat_map_inode`: {fa}')
        return (
            self.flat_map_map(fa, f)
            if isinstance(fa, MapNode) else
            self.flat_map_list(fa, f)
            if isinstance(fa, ListNode) else
            err()
        )

    def flat_map_map(self, fa: MapNode[A], f: Callable[[A], Node[B, Any]]) -> Node[B, Any]:
        return MapNode(fa.sub.valmap(lambda a: self.flat_map(a, f)))

    def flat_map_list(self, fa: ListNode[A], f: Callable[[A], Node[B, Any]]) -> Node[B, Any]:
        return ListNode(fa.sub.map(lambda a: self.flat_map(a, f)))

    def flat_map_leaf(self, fa: LeafNode[A], f: Callable[[A], Node[B, Any]]) -> Node[B, Any]:
        return f(fa.data)

    def map(self, fa: Node[A, Any], f: Callable[[A], B]) -> Node[B, Any]:
        return (
            self.map_inode(fa, f)
            if isinstance(fa, Inode) else
            self.map_leaf(fa, f)
        )

    def map_inode(self, fa: Inode[A, Any], f: Callable[[A], B]) -> Node[B, Any]:
        def err() -> Inode[A, Any]:
            raise Exception(f'invalid sub for `TreeFlatMap.map_inode`: {fa}')
        return (
            self.map_map(fa, f)
            if isinstance(fa, MapNode) else
            self.map_list(fa, f)
            if isinstance(fa, ListNode) else
            err()
        )

    def map_map(self, fa: MapNode[A], f: Callable[[A], B]) -> Node[B, Any]:
        return MapNode(fa.data.valmap(lambda a: self.map(a, f)))

    def map_list(self, fa: ListNode[A], f: Callable[[A], B]) -> Node[B, Any]:
        return ListNode(fa.sub.map(lambda a: self.map(a, f)))

    def map_leaf(self, fa: LeafNode[A], f: Callable[[A], B]) -> Node[B, Any]:
        return LeafNode(f(fa.data))


class SubTree(Implicits, implicits=True, auto=True):

    @staticmethod
    def cons(fa: Node, key: Key) -> 'SubTree':
        return (  # type: ignore
            cast(SubTree, SubTreeList(fa, key))
            if isinstance(fa, ListNode) else
            SubTreeLeaf(fa, key)
            if isinstance(fa, LeafNode) else
            SubTreeMap(fa, key)
        )

    @staticmethod
    def from_maybe(data: Maybe[Node], key: Key, err: str) -> 'SubTree':
        return data.cata(SubTree.cons, SubTreeInvalid(key, err))

    def __getattr__(self, key: Key) -> 'SubTree':
        try:
            return super().__getattr__(key)
        except AttributeError:
            return self._getattr(key)

    @abc.abstractmethod
    def _getattr(self, key: Key) -> 'SubTree':
        ...

    def __getitem__(self, key: Key) -> 'SubTree':
        return self._getitem(key)

    @abc.abstractmethod
    def _getitem(self, key: Key) -> 'SubTree':
        ...

    def cata(self, f: Callable[[Node], A], b: Union[A, Callable[[], A]]) -> A:
        return (
            f(self.data)
            if isinstance(self, SubTreeValid)
            else call_by_name(b)
        )

    @abc.abstractproperty
    def e(self) -> Either[str, Node]:
        ...

    @abc.abstractproperty
    def valid(self) -> Boolean:
        ...

    @abc.abstractproperty
    def strings(self) -> LazyList[str]:
        ...

    @abc.abstractproperty
    def show(self) -> LazyList[str]:
        ...

    @property
    def rule(self) -> Either[str, str]:
        return self.e.map(_.rule)


class SubTreeValid(SubTree):

    def __init__(self, data: Node, key: Key) -> None:
        self.data = data
        self._key = key

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.data)

    @property
    def e(self) -> Either[str, Node]:
        return Right(self.data)

    @property
    def valid(self) -> Boolean:
        return true

    @property
    def strings(self) -> LazyList[str]:
        return self.data.strings

    @property
    def show(self) -> str:
        return self.data.show


class SubTreeList(SubTreeValid):

    @property
    def head(self) -> SubTree:
        return self[0]

    @property
    def last(self) -> SubTree:
        return self[-1]

    def _getattr(self, key: Key) -> SubTree:
        return SubTreeInvalid(key, 'cannot access attrs in SubTreeList')

    def _getitem(self, key: Key) -> SubTree:
        return self.data.lift(key)

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.data.sub_l.drain.join_comma)

    @property
    def _keys(self) -> LazyList[str]:
        return self.data.k


class SubTreeLeaf(SubTreeValid):

    def err(self, key: Key) -> SubTree:
        return SubTreeInvalid(key, 'cannot access attrs in SubTreeLeaf')

    def _getattr(self, key: Key) -> SubTree:
        return self.err(key)

    def _getitem(self, key: Key) -> SubTree:
        return self.err(key)


class SubTreeMap(SubTreeValid):

    def _getattr(self, key: Key) -> SubTree:
        return self.data.lift(key)

    def _getitem(self, key: Key) -> SubTree:
        return self.data.lift(key)

    @property
    def _keys(self) -> LazyList[str]:
        return self.data.k


class SubTreeInvalid(SubTree):

    def __init__(self, key: Key, reason: str) -> None:
        self.key = key
        self.reason = reason

    def __str__(self) -> str:
        s = 'SubTreeInvalid({}, {})'
        return s.format(self.key, self.reason)

    def __repr__(self) -> str:
        return str(self)

    @property
    def valid(self) -> Boolean:
        return false

    @property
    def _error(self) -> str:
        return 'no subtree `{}`: {}'.format(self.key, self.reason)

    def _getattr(self, key: Key) -> SubTree:
        return self

    def _getitem(self, key: Key) -> SubTree:
        return self

    @property
    def e(self) -> Either[str, Node]:
        return Left(self._error)

    @property
    def strings(self) -> LazyList[str]:
        return LazyList([])

    @property
    def show(self) -> LazyList[str]:
        return str(self)

__all__ = ('Node', 'Inode', 'LeafNode')



from typing import Any

from amino import options
from amino.util.mod import unsafe_import_name

_ = None
__ = None
L = None


def set(mod: str) -> None:
    def name(name: str) -> Any:
        return unsafe_import_name(mod, name)
    global _, __, L
    _ = name('AttrLambdaInst')
    __ = name('MethodLambdaInst')
    L = name('ComplexLambdaInit')


def set_debug() -> None:
    set('amino.anon.debug')


def set_prod() -> None:
    set('amino.anon.prod')

if options.anon_debug:
    set_debug()
else:
    set_prod()

__all__ = ('set_debug', 'set_prod', '_', '__', 'L')



import abc
from typing import TypeVar, Generic, Callable
import operator

from lenses import lens, Lens

from amino.tc.base import TypeClass, tc_prop
from amino.tc.functor import Functor
from amino.func import curried, I
from amino.maybe import Maybe, Empty, Just
from amino.boolean import Boolean
from amino import _
from amino.tc.monoid import Monoid
from amino.tc.monad import Monad

G = TypeVar('G')
H = TypeVar('H')
A = TypeVar('A')
B = TypeVar('B')
Z = TypeVar('Z')


class FoldableABC(Generic[A], abc.ABC):
    pass

F = FoldableABC


class Foldable(Generic[H], TypeClass[H]):
    # FIXME lens functions return index lenses, which is not a property of
    # Foldable

    @abc.abstractmethod
    def with_index(self, fa: F[A]) -> F[A]:
        ...

    @abc.abstractmethod
    def filter(self, fa: F[A], f: Callable[[A], bool]) -> F:
        ...

    def filter_not(self, fa: F[A], f: Callable[[A], bool]) -> F:
        pred = lambda a: not f(a)
        return self.filter(fa, pred)

    def filter_type(self, fa: F[A], tpe: type) -> F:
        pred = lambda a: isinstance(a, tpe)
        return self.filter(fa, pred)

    @abc.abstractmethod
    def find(self, fa: F[A], f: Callable[[A], bool]) -> Maybe[A]:
        ...

    @abc.abstractmethod
    @curried
    def fold_left(self, fa: F[A], z: Z, f: Callable[[Z, A], Z]) -> Z:
        ...

    def fold_map(self, fa: F[A], z: B, f: Callable[[A], B], g: Callable[[Z, B], Z]=operator.add) -> Z:
        ''' map `f` over the traversable, then fold over the result
        using the supplied initial element `z` and operation `g`,
        defaulting to addition for the latter.
        '''
        mapped = Functor.fatal(type(fa)).map(fa, f)
        return self.fold_left(mapped)(z)(g)

    @curried
    def fold_m(self, fa: F[A], z: G, f: Callable[[B, A], G]) -> G:
        monad = Monad.fatal(type(z))
        def folder(z1: G, a: A) -> G:
            return monad.flat_map(z1, lambda b: f(b, a))
        return self.fold_left(fa)(z)(folder)

    def fold(self, fa: F[A], tpe: type) -> A:
        mono = Monoid.fatal(tpe)
        return self.fold_left(fa)(mono.empty)(mono.combine)

    @abc.abstractmethod
    def find_map(self, fa: F[A], f: Callable[[A], Maybe[B]]) -> Maybe[B]:
        ...

    def find_type(self, fa: F[A], tpe: type) -> Maybe[A]:
        pred = lambda a: isinstance(a, tpe)
        return self.find(fa, pred)

    @abc.abstractmethod
    def index_where(self, fa: F[A], f: Callable[[A], bool]) -> Maybe[int]:
        ...

    def index_of(self, fa: F[A], a: A) -> Maybe[int]:
        return self.index_where(fa, _ == a)

    def exists(self, fa: F[A], f: Callable[[A], bool]) -> Boolean:
        return Boolean(self.find(fa, f).is_just)

    def contains(self, fa: F[A], v: A) -> Boolean:
        return self.exists(fa, _ == v)

    def lens(self, fa: F[A], f: Callable[[A], bool]) -> Maybe[Lens]:
        return self.index_where(fa, f) / (lambda i: lens()[i])

    def find_lens(self, fa: F[A], f: Callable[[A], Maybe[Lens]]) -> Maybe[Lens]:
        check = lambda a: f(a[1]) / (lambda b: (a[0], b))
        index = lambda i, l: lens()[i].add_lens(l)
        wi = self.with_index(fa)
        return self.find_map(wi, check).map2(index)

    def find_lens_pred(self, fa: F[A], f: Callable[[A], bool]) -> Maybe[Lens]:
        g = lambda a: Boolean(f(a)).maybe(lens())
        return self.find_lens(fa, g)

    def _min_max(self, fa: F[A], f: Callable[[A], int], pred: Callable[[int, int], int]) -> Maybe[int]:
        def folder(z: Maybe[A], a: A) -> Maybe[int]:
            return (
                z.map(lambda b: b if pred(f(b), f(a)) else a)
                .or_else(Just(a))
            )
        return self.fold_left(fa, Empty())(folder)

    def max_by(self, fa: F[A], f: Callable[[A], int]) -> Maybe[A]:
        return self._min_max(fa, f, operator.gt)

    def min_by(self, fa: F[A], f: Callable[[A], int]) -> Maybe[A]:
        return self._min_max(fa, f, operator.lt)

    @tc_prop
    def max(self, fa: F[A]) -> Maybe[A]:
        return self.max_by(fa, I)

    @tc_prop
    def min(self, fa: F[A]) -> Maybe[A]:
        return self.min_by(fa, I)

__all__ = ('Foldable',)



from .path import (setup, create_temp_file, temp_file, temp_path, temp_dir,
                   fixture_path, load_fixture)

__all__ = ('create_temp_file', 'temp_file', 'temp_path', 'temp_dir',
           'fixture_path', 'load_fixture', 'setup')



import os

import amino


class EnvOption:

    def __init__(self, name: str) -> None:
        self.name = name

    def __bool__(self) -> bool:
        return self.name in os.environ

    @property
    def value(self) -> 'amino.Maybe[str]':
        return amino.env[self.name]

    def __str__(self) -> str:
        value = self.value.map(lambda a: f'={a}') | ' is unset'
        return f'{self.name}{value}'

development = EnvOption('AMINO_DEVELOPMENT')
integration_test = EnvOption('AMINO_INTEGRATION')
anon_debug = EnvOption('AMINO_ANON_DEBUG')

__all__ = ('development', 'integration_test', 'anon_debug')



from typing import Generic, TypeVar, Callable, Any

from amino.tc.base import Implicits
from amino.tc.monad import Monad
from amino.util.string import ToStr
from amino import List

A = TypeVar('A')
B = TypeVar('B')


class Id(Generic[A], Implicits, ToStr, implicits=True, auto=True):

    def __init__(self, value: A) -> None:
        self.value = value

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Id) and self.value == other.value

    @property
    def _arg_desc(self) -> List[str]:
        return List(str(self.value))


class IdMonad(Monad[Id], tpe=Id):

    def pure(self, a: A) -> Id[A]:
        return Id(a)

    def flat_map(self, fa: Id[A], f: Callable[[A], Id[B]]) -> Id[B]:
        return f(fa.value)


__all__ = ('Id',)



from typing import Callable, Tuple, Any
from types import FunctionType

from amino.anon.prod.attr import AttrLambdaInst
from amino.anon.prod.method import Anon, MethodRef, MethodLambdaInst, MethodChain, AnonChain
from amino.util.fun import format_funcall


def make_complex(args: list) -> Tuple[list, list, int, Any, int]:
    i_strict, i_lambda, i_param = 0, 0, 0
    ordered = ''
    stricts = ''
    lambdas = ''
    params = ''
    rest = []
    lambda_args = []
    for a in args:
        if a is AttrLambdaInst:
            i_param += 1
            arg = f'y{i_param},'
            ordered += arg
            params += arg
        elif isinstance(a, Anon):
            i_param += 1
            i_lambda += 1
            arg = f'y{i_param},'
            ordered += f'l{i_lambda}.__substitute_object__(y{i_param}),'
            params += arg
            lambdas += f'l{i_lambda},'
            lambda_args.append(a)
        else:
            i_strict += 1
            arg = f'x{i_strict},'
            ordered += arg
            stricts += arg
            rest.append(a)
    rep = f'lambda f: lambda {lambdas}: lambda {params}: lambda {stricts}**kw: f({ordered}**kw)'
    return lambda_args, rest, rep, eval(rep), i_param


class ComplexLambda:

    def __init__(self, func: Callable, a: tuple, kw: dict) -> None:
        self.__func = func
        self.__args = a
        self.__kwargs = kw
        self.__name__ = (
            self.__func.__name__
            if isinstance(self.__func, FunctionType) else
            self.__func.__class__.__name__
        )
        self.__qualname__ = self.__name__
        self.__annotations__ = {}
        self.__lambda_args, self.__rest, self.__repr, self.__lambda, self.__param_count = make_complex(self.__args)

    def __call__(self, *a, **kw):
        return self.__lambda(self.__func)(*self.__lambda_args)(*a)(*self.__rest, **self.__kwargs, **kw)

    def __readable(self) -> str:
        return format_funcall(self.__func, self.__args, self.__kwargs)

    def __str__(self) -> str:
        return self.__repr

    def __rshift__(self, f):
        return AnonChain(self, f, self.__param_count)

    def __getattr__(self, name):
        return MethodChain(self, f'a.{name}')


class LazyMethod(Anon):

    def __init__(self, obj, attr: MethodRef) -> None:
        self.__obj = obj
        self.__attr = attr
        self.__name__ = self.__attr.__name__

    def __call__(self, *a, **kw):
        return self.__attr(*a, **kw)(self.__obj)

    def __getattr__(self, name: str):
        return LazyMethod(self.__obj, getattr(self.__attr, name))


class ComplexLambdaInit:

    def __init__(self, func) -> None:
        self.__func = func

    def __call__(self, *a, **kw):
        return ComplexLambda(self.__func, a, kw)

    def __getattr__(self, name):
        return ComplexLambdaInit(LazyMethod(self.__func, getattr(MethodLambdaInst, name)))

__all__ = ('ComplexLambdaInit',)



import pathlib

from amino.typecheck import boot
from amino.maybe import Maybe, Just, Empty, may, flat_may, Nothing
from amino.either import Left, Right, Either
from amino.list import List, Lists
from amino.boolean import Boolean
from amino.anon.boot import __, L, _
from amino.lazy_list import LazyList
from amino.map import Map
from amino.future import Future
from amino.func import curried, I
from amino.env_vars import env
from amino.task import Try, Task, IO
from amino.logging import Logger, log
from amino.eff import Eff
from amino.eval import Eval
from amino.regex import Regex
from amino.options import integration_test, development
from amino.do import do
from amino.id import Id

Path = pathlib.Path

boot()

__all__ = ('Maybe', 'Just', 'Empty', 'may', 'List', 'Map', '_', 'Future', 'Boolean', 'development', 'flat_may',
           'curried', '__', 'Left', 'Right', 'Either', 'env', 'Try', 'LazyList', 'Logger', 'log', 'I', 'L', 'Eff',
           'Task', 'Eval', 'Regex', 'Nothing', 'integration_test', 'Lists', 'do', 'IO', 'Id')



import sys
from pathlib import Path

from amino import env

__base_dir__ = None


class TestEnvError(Exception):
    pass


def setup(path):
    ''' Use the supplied path to initialise the tests base dir.
    If _path is a file, its dirname is used.
    '''
    if not isinstance(path, Path):
        path = Path(path)
    if not path.is_dir():
        path = path.parent
    global __base_dir__
    __base_dir__ = path
    container = str(pkg_dir())
    if container not in sys.path:
        sys.path.insert(0, container)
        env['PYTHONPATH'] = '{}:{}'.format(container, env['PYTHONPATH'] | '')


def _check():
    if __base_dir__ is None:
        msg = 'Test base dir not set! Call amino.test.setup(dir).'
        raise TestEnvError(msg)


def temp_path(*components):
    _check()
    return Path(__base_dir__, '_temp', *components)


def temp_dir(*components):
    _dir = temp_path(*components)
    _dir.mkdir(exist_ok=True, parents=True)
    return _dir


def temp_file(*components):
    return temp_dir(*components[:-1]).joinpath(*components[-1:])


def create_temp_file(*components):
    _file = temp_file(*components)
    _file.touch()
    return _file


def fixture_path(*components):
    _check()
    return Path(__base_dir__, '_fixtures', *components)


def load_fixture(*components):
    with fixture_path(*components).open() as f:
        return f.read()


def base_dir():
    return __base_dir__


def pkg_dir() -> Path:
    return base_dir().parent

__all__ = ('create_temp_file', 'temp_file', 'temp_path', 'temp_dir',
           'fixture_path', 'load_fixture', 'base_dir', 'pkg_dir')



from typing import TypeVar, Generic

from amino.tc.base import TypeClass

A = TypeVar('A')


class Show(Generic[A], TypeClass[A]):

    def show(self, obj):
        return str(obj)

__all__ = ('Show',)



from typing import Callable, TypeVar

from amino import Just, L, _, Empty
from amino.tc.monad import Monad
from amino.tc.base import ImplicitInstances
from amino.lazy import lazy
from amino.task import Task
from amino.util.fun import lambda_str

A = TypeVar('A')
B = TypeVar('B')


class TaskInstances(ImplicitInstances):

    @lazy
    def _instances(self):
        from amino.map import Map
        return Map({Monad: TaskMonad()})


class TaskMonad(Monad):

    def pure(self, a: A):
        return Task.now(a)

    def flat_map(self, fa: Task[A], f: Callable[[A], Task[B]]) -> Task[B]:
        return fa.flat_map(f)

    def map(self, fa: Task[A], f: Callable[[A], B]) -> Task[B]:
        s = 'map({})'.format(lambda_str(f))
        mapper = L(f)(_) >> L(Task.now)(_)
        return fa.flat_map(mapper, Empty(), Just(s))

__all__ = ('TaskInstances',)



from typing import TypeVar, Generic, Callable, Any

from amino.tc.base import TypeClass

A = TypeVar('A')
B = TypeVar('B')


class Tap(Generic[A], TypeClass):

    def tap(self, a: A, f: Callable[[A], Any]) -> A:
        f(a)
        return a

__all__ = ('Tap',)



from traceback import format_tb

from amino import List, Lists


def sanitize_tb(tb: List[str]) -> List[str]:
    return tb.flat_map(lambda a: Lists.wrap(a.splitlines()))


def format_exception(exc: Exception) -> List[str]:
    tb = sanitize_tb(Lists.wrap(format_tb(exc.__traceback__)))
    return tb.cat(f'{exc.__class__.__name__}: {exc}')

__all__ = ('sanitize_tb', 'format_exception')



from amino.test.sure_ext import install_assertion_builder, AssBuilder


class SureSpec:

    def setup(self) -> None:
        install_assertion_builder(AssBuilder)  # type: ignore

__all__ = ('SureSpec',)



import abc
import time
import traceback
import inspect
from typing import Callable, TypeVar, Generic, Any, Coroutine

from fn.recur import tco

from amino import Either, Right, Left, Maybe, List, Empty, __, Just, env, _, Lists, L
from amino.tc.base import Implicits, ImplicitsMeta
from amino.logging import log
from amino.util.fun import lambda_str, format_funcall
from amino.util.exception import format_exception, sanitize_tb

A = TypeVar('A')
B = TypeVar('B')


class TaskException(Exception):
    remove_pkgs = List('amino', 'fn')

    def __init__(self, f, stack, cause) -> None:
        self.f = f
        self.stack = List.wrap(stack)
        self.cause = cause

    @property
    def location(self):
        files = List('task', 'anon', 'instances/task', 'tc/base')
        def filt(entry, name):
            return entry.filename.endswith('/amino/{}.py'.format(name))
        stack = self.stack.filter_not(lambda a: files.exists(L(filt)(a, _)))
        pred = (lambda a: not TaskException.remove_pkgs
                .exists(lambda b: '/{}/'.format(b) in a.filename))
        return stack.find(pred)

    @property
    def format_stack(self) -> List[str]:
        rev = self.stack.reversed
        def remove_recursion(i):
            pre = rev[:i + 1]
            post = rev[i:].drop_while(__.filename.endswith('/amino/task.py'))
            return pre + post
        def remove_internal():
            start = rev.index_where(_.function == 'unsafe_perform_sync')
            return start / remove_recursion | rev
        frames = (self.location.to_list if Task.stack_only_location else remove_internal())
        data = frames / (lambda a: a[1:-2] + tuple(a[-2]))
        return sanitize_tb(Lists.wrap(traceback.format_list(list(data))))

    @property
    def lines(self) -> List[str]:
        cause = format_exception(self.cause)
        suf1 = '' if self.stack.empty else ' at:'
        tb1 = (List() if self.stack.empty else self.format_stack)
        return tb1.cons(f'Task exception{suf1}').cat('Cause:') + cause + List(
            '',
            'Callback:',
            f'  {self.f}'
        )

    def __str__(self):
        return self.lines.join_lines


class TaskMeta(ImplicitsMeta):

    @property
    def zero(self):
        return Task.now(None)


class Task(Generic[A], Implicits, implicits=True, metaclass=TaskMeta):
    debug = 'AMINO_TASK_DEBUG' in env
    stack_only_location = True

    def __init__(self) -> None:
        self.stack = inspect.stack() if Task.debug else []

    @staticmethod
    def delay(f: Callable[..., A], *a, **kw):
        try:
            s = format_funcall(f, a, kw)
        except Exception as e:
            s = str(f)
            log.error(e)
        return Suspend(L(f)(*a, **kw) >> Now, s)

    @staticmethod
    def suspend(f: Callable[..., 'Task[A]'], *a, **kw):
        try:
            s = format_funcall(f, a, kw)
        except Exception as e:
            s = str(f)
            log.error(e)
        return Suspend(L(f)(*a, **kw), s)

    @staticmethod
    def call(f: Callable[..., A], *a, **kw):
        return Task.delay(f, *a, **kw)

    @staticmethod
    def now(a: A) -> 'Task[A]':
        return Now(a)

    @staticmethod
    def pure(a: A) -> 'IO[A]':
        return Pure(a)

    @staticmethod
    def just(a: A) -> 'Task[Maybe[A]]':
        return Task.now(Just(a))

    @staticmethod
    def failed(err: str) -> 'Task[A]':
        def fail():
            raise Exception(err)
        return Task.suspend(fail)

    @staticmethod
    def from_either(a: Either[Any, A]) -> 'Task[A]':
        return a.cata(Task.failed, Task.now)

    @staticmethod
    def from_maybe(a: Maybe[A], error: str) -> 'Task[A]':
        return a / Task.now | Task.failed(error)

    def run(self):
        @tco
        def run(t):
            if isinstance(t, Now):
                return True, (t.value,)
            elif isinstance(t, Task):
                return True, (t.step(),)
            else:
                return False, t
        return run(self)

    @property
    def _name(self):
        return self.__class__.__name__

    def flat_map(self, f: Callable[[A], 'Task[B]'], ts=Empty(), fs=Empty()
                 ) -> 'Task[B]':
        ts = ts | (lambda: self.string)
        fs = fs | (lambda: 'flat_map({})'.format(lambda_str(f)))
        return self._flat_map(f, ts, fs)

    @abc.abstractmethod
    def _flat_map(self, f: Callable[[A], 'Task[B]'], ts, fs) -> 'Task[B]':
        ...

    def step(self):
        try:
            return self._step_timed() if Task.debug else self._step()
        except TaskException as e:
            raise e
        except Exception as e:
            raise TaskException(self.string, self.stack, e)

    @abc.abstractmethod
    def _step(self):
        ...

    def _step_timed(self):
        start = time.time()
        v = self._step()
        dur = time.time() - start
        if dur > 0.1:
            log.ddebug(lambda: 'task {} took {:.4f}s'.format(self.string, dur))
        return v

    def __repr__(self):
        return 'Task({})'.format(self.string)

    @property
    def attempt(self) -> Either[Exception, A]:
        try:
            return Right(self.run())
        except TaskException as e:
            return Left(e)

    def unsafe_perform_sync(self):
        return self.attempt

    @property
    def fatal(self) -> A:
        return self.attempt.get_or_raise

    def and_then(self, nxt: 'Task[B]'):
        fs = 'and_then({})'.format(nxt.string)
        return self.flat_map(lambda a: nxt, fs=Just(fs))

    __add__ = and_then

    def join_maybe(self, err):
        return self.flat_map(lambda a: Task.from_maybe(a, err))

    @property
    def join_either(self):
        return self.flat_map(Task.from_either)

    def with_string(self, s):
        self.string = s
        return self

    def with_stack(self, s):
        self.stack = s
        return self

    def recover(self, f: Callable[[TaskException], B]) -> 'Task[B]':
        return Task.delay(self.unsafe_perform_sync).map(__.value_or(f))

    @property
    def coro(self) -> Coroutine:
        async def coro() -> Either[TaskException, A]:
            return self.attempt
        return coro()


class Suspend(Generic[A], Task[A]):

    def __init__(self, thunk: Callable, string) -> None:
        super().__init__()
        self.thunk = thunk
        self.string = string

    def _step(self):
        return self.thunk().with_stack(self.stack)

    def __str__(self):
        return '{}({})'.format(self._name, self.string)

    def _flat_map(self, f: Callable[[A], 'Task[B]'], ts, fs) -> 'Task[B]':
        return BindSuspend(self.thunk, f, ts, fs)


class BindSuspend(Generic[A], Task[A]):

    def __init__(self, thunk, f: Callable, ts, fs) -> None:
        super().__init__()
        self.thunk = thunk
        self.f = f
        self.ts = ts
        self.fs = fs

    def _step(self):
        return (
            self.thunk()
            .flat_map(self.f, fs=Just(self.fs))
            .with_stack(self.stack)
        )

    def __str__(self):
        return '{}({})'.format(self._name, self.string)

    def _flat_map(self, f: Callable[[A], 'Task[B]'], ts, fs) -> 'Task[B]':
        bs = L(BindSuspend)(self.thunk, lambda a: self.f(a).flat_map(f, Just(ts), Just(fs)), ts, fs)
        return Suspend(bs, '{}.{}'.format(ts, fs))

    @property
    def string(self):
        return '{}.{}'.format(self.ts, self.fs)


class Now(Generic[A], Task[A]):

    def __init__(self, value) -> None:
        super().__init__()
        self.value = value

    def _step(self):
        return self

    def __str__(self):
        return '{}({})'.format(self._name, self.value)

    def _flat_map(self, f: Callable[[A], 'Task[B]'], ts, fs) -> 'Task[B]':
        return Suspend(L(f)(self.value), '{}.{}'.format(ts, fs))

    @property
    def ts(self):
        return 'Task'

    @property
    def fs(self):
        return 'now({})'.format(self.value)

    @property
    def string(self) -> str:
        return str(self)

Pure = Now


def Try(f: Callable[..., A], *a, **kw) -> Either[Exception, A]:
    return Task.delay(f, *a, **kw).attempt


def task(fun):
    def dec(*a, **kw):
        return Task.delay(fun, *a, **kw)
    return dec

IO = Task

__all__ = ('Task', 'Try', 'task', 'IO')



import abc
from typing import Generic, TypeVar, Callable, Tuple, Any

from amino import LazyList, Maybe, I, Empty, Just, Boolean
from amino.lazy import lazy
from amino.lazy_list import LazyLists
from amino.tree import indent, Node
from amino.boolean import true, false


Data = TypeVar('Data')
A = TypeVar('A')
B = TypeVar('B')


class RoseTree(Generic[Data]):

    @abc.abstractproperty
    def parent(self) -> 'RoseTree':
        ...

    @abc.abstractproperty
    def is_root(self) -> Boolean:
        ...

    def __init__(self, data: Data, sub_cons: Callable[['RoseTree[Data]'], LazyList['RoseTree[Data]']]) -> None:
        self.data = data
        self._sub_cons = sub_cons

    @lazy
    def sub(self) -> LazyList['RoseTree[Data]']:
        return self._sub_cons(self)

    @property
    def desc(self) -> str:
        num = self.sub._drain().length
        return f'{self.data}, {num} children'

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.desc})'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.data})'

    @property
    def show(self) -> str:
        return self.strings._drain().mk_string('\n')

    @property
    def strings(self) -> LazyList[str]:
        return self._strings()

    def _strings(self) -> LazyList[str]:
        return indent(self.sub.flat_map(lambda a: a._strings())).cons(str(self.data))

    def map(self, f: Callable[[Data], A]) -> 'RoseTree[A]':
        return self.copy(f, I)

    def copy(self, f: Callable[[Data], A], g: Callable[[LazyList['RoseTree[Data]']], LazyList['RoseTree[A]']]
             ) -> 'RoseTree[A]':
        return RoseTreeRoot(f(self.data), self._copy_sub(f, g))

    def _copy_sub(self, f: Callable[[Data], A], g: Callable[[LazyList['RoseTree[Data]']], LazyList['RoseTree[A]']]
                  ) -> Callable[['RoseTree[A]'], LazyList['RoseTree[A]']]:
        def sub(node: RoseTree[Data]) -> Callable[[RoseTree[A]], RoseTree[A]]:
            return lambda parent: BiRoseTree(f(node.data), parent, node._copy_sub(f, g))
        return lambda parent: g(self.sub).map(lambda a: sub(a)(parent))

    def __getitem__(self, key: int) -> Maybe['RoseTree[Data]']:
        return self.sub.lift(key)

    def filter(self, f: Callable[[Data], bool]) -> 'RoseTree[Data]':
        return self.copy(I, lambda a: a.filter(lambda n: f(n.data)))

    def _drain(self) -> None:
        self.sub._drain().foreach(lambda a: a._drain())


class RoseTreeRoot(Generic[Data], RoseTree[Data]):

    @property
    def parent(self) -> RoseTree[Data]:
        return self

    @property
    def is_root(self) -> Boolean:
        return true


class BiRoseTree(Generic[Data], RoseTree[Data]):

    def __init__(
            self,
            data: Data,
            parent: RoseTree[Data],
            sub_cons: Callable[[RoseTree[Data]], LazyList[RoseTree[Data]]]
    ) -> None:
        super().__init__(data, sub_cons)
        self._parent = parent

    @property
    def parent(self) -> RoseTree[Data]:
        return self._parent

    @property
    def is_root(self) -> Boolean:
        return false


def node(data: Data, sub_cons: Callable[[RoseTree[Data]], LazyList[RoseTree[Data]]]
         ) -> Callable[[RoseTree[Data]], RoseTree[Data]]:
    return lambda parent: BiRoseTree(data, parent, sub_cons)


def nodes(*s: Tuple[Data, Callable[[RoseTree[Data]], RoseTree[Data]]]
          ) -> Callable[[RoseTree[Data]], LazyList[RoseTree[Data]]]:
    return lambda parent: LazyList(s).map2(node).map(lambda f: f(parent))


def leaf(data: Data) -> Callable[[RoseTree[Data]], RoseTree[Data]]:
    return lambda parent: BiRoseTree(data, parent, lambda a: LazyLists.empty())


def leaves(*data: Data) -> Callable[[RoseTree[Data]], LazyList[RoseTree[Data]]]:
    return lambda parent: LazyList(data).map(leaf).map(lambda f: f(parent))


def from_tree(
        tree: Node[A, Any],
        f: Callable[[Node[A, Any], Maybe[RoseTree[B]]], B],
        cons_root: Callable[[B, Callable[[RoseTree[Node[A, Any]]], LazyList[RoseTree[Node[A, Any]]]]], RoseTree[B]],
        cons_node: Callable[
            [B, RoseTree[Node[A, Any]], Callable[[RoseTree[Node[A, Any]]], LazyList[RoseTree[Node[A, Any]]]]],
            RoseTree[B]
        ]
) -> RoseTree[B]:
    def sub(node: Node[A, Any]) -> Callable[[RoseTree[Node[B, Any]]], LazyList[RoseTree[Node[B, Any]]]]:
        def cons_sub(parent: RoseTree[Node[B, Any]], a: A) -> RoseTree[B]:
            return cons_node(f(a, Just(parent)), parent, sub(a))
        return lambda parent: node.sub_l.map(lambda a: cons_sub(parent, a))
    return cons_root(f(tree, Empty()), sub(tree))


def from_tree_default(tree: Node[A, Any], f: Callable[[Node[A, Any], Maybe[RoseTree[B]]], B]) -> RoseTree[B]:
    return from_tree(tree, f, RoseTreeRoot, BiRoseTree)

__all__ = ('BiRoseTree', 'RoseTree', 'node', 'leaf', 'leaves', 'from_tree')
